from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.generate.exec_param import ExecParam


async def should_operate(hub, execution_context: ExecutionContext) -> bool:
    return (
        execution_context.get_exec_param_value(ExecParam.PROCESS_RESOURCE_ID_PARAM)
        and execution_context.get_method_param_value("resource_id") is not None
    )


async def operate(hub, execution_context: ExecutionContext) -> None:
    extended_params = {**execution_context.method_params}
    resource_id = extended_params.pop("resource_id", None)
    if resource_id is not None:
        resource_id_params = (
            hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
                execution_context.resource_type, resource_id
            )
        )
        aliases = execution_context.exec_params[ExecParam.PATH_PARAM_ALIASES]
        if not aliases:
            extended_params.update(resource_id_params)
        else:
            for resource_id_param in resource_id_params:
                if resource_id_param in aliases:
                    extended_params[aliases[resource_id_param]] = resource_id_params[
                        resource_id_param
                    ]

    execution_context.method_params = extended_params



RESOURCE_TYPE_FULL = f"gcp.storage.bucket"


async def enforce_retention_policy(
    hub,
    ctx,
    result,
    old_is_locked: bool = None,
    new_is_locked: bool = None,
    if_metageneration_match: str = None,
    user_project: str = None,
    resource_id: str = None,
):
    name = result.get("name")

    if new_is_locked is None:
        # not specified in SLS -> no update
        return

    if old_is_locked is None:
        # default GCP value
        old_is_locked = False

    old_is_locked = bool(old_is_locked)
    new_is_locked = bool(new_is_locked)

    if old_is_locked == new_is_locked:
        # no update needed
        return

    if old_is_locked and not new_is_locked:
        result["result"] = False
        result["comment"].append(
            f"Invalid modification of retention_policy.is_locked for {RESOURCE_TYPE_FULL} '{name}' - Retention policy cannot be unlocked once locked"
        )
        return

    if if_metageneration_match is None:
        result["result"] = False
        result["comment"].append(
            f"Locking retention policy for {RESOURCE_TYPE_FULL} '{name}' requires if_metageneration_match SLS property"
        )
        return

    if ctx["test"]:
        result["comment"].append(
            f"Would lock retention policy for {RESOURCE_TYPE_FULL} '{name}'."
        )
        if result["new_state"]["retention_policy"]:
            result["new_state"]["retention_policy"]["is_locked"] = True
        else:
            result["new_state"]["retention_policy"] = {"is_locked": True}
        result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
            result["new_state"]
        )
        return
    else:
        lock_ret = await hub.exec.gcp.storage.bucket.lock_retention_policy(
            ctx=ctx,
            if_metageneration_match=if_metageneration_match,
            resource_id=resource_id,
            user_project=user_project,
        )
        if not lock_ret["result"]:
            result["result"] = False
            result["comment"].append(
                f"Failed to lock retention policy for {RESOURCE_TYPE_FULL} '{name}'."
            )
            result["comment"] += lock_ret["comment"]
            return
        result["comment"].append(
            f"Retention policy locked for {RESOURCE_TYPE_FULL} '{name}'."
        )



from typing import Any
from typing import Dict
from typing import List


async def update_named_ports(
    hub,
    ctx,
    current_instance_group: Dict[str, Any],
    new_named_ports: List[Dict[str, Any]],
) -> Dict[str, Any]:
    # Update the updatable named port properties
    result = {"result": False, "comment": []}

    ret = await hub.exec.gcp.compute.instance_group.set_named_ports(
        ctx,
        current_instance_group["resource_id"],
        new_named_ports,
        current_instance_group["fingerprint"],
    )

    if not ret["result"] or not ret["ret"]:
        result["comment"] += ret["comment"]
        return result

    result["result"] = True
    result["ret"] = ret["ret"]

    return result



"""State module for managing target tools."""
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    region: str = None,
    project: str = None,
    instances: List[str] = None,
    session_affinity: str = None,
    description: str = None,
    failover_ratio: float = None,
    health_checks: List[str] = None,
    backup_pool: str = None,
    request_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Creates a target pool in the specified project and region using the data included in the request.

    Args:
        name (str):
            An Idem name of the resource.
        instances (List[str], Optional):
            A list of resource URLs to the virtual machine instances serving this pool. They must live in zones contained in the same region as this pool. Defaults to None.
        session_affinity (str, Optional):
            Session affinity option, must be one of the following values:
            NONE: Connections from the same client IP may go to any instance in the pool.
            CLIENT_IP: Connections from the same client IP will go to the same instance in the pool while that instance remains healthy.
            CLIENT_IP_PROTO: Connections from the same client IP with the same IP protocol will go to the same instance in the pool while that instance remains healthy.
        description (str, Optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.
        region(str, Optional):
            Name of the region scoping this request.
        failover_ratio(float, Optional):
            This field is applicable only when the containing target pool is serving a forwarding rule as the primary pool (i.e., not as a backup pool to some other target pool). The value of the field must be in [0, 1]. If set, backupPool must also be set. They together define the fallback behavior of the primary target pool: if the ratio of the healthy instances in the primary pool is at or below this number, traffic arriving at the load-balanced IP will be directed to the backup pool. In case where failoverRatio is not set or all the instances in the backup pool are unhealthy, the traffic will be directed back to the primary pool in the "force" mode, where traffic will be spread to the healthy instances with the best effort, or to all instances when no instance is healthy. Defaults to None.
        health_checks(List[str], Optional):
            The URL of the HttpHealthCheck resource. A member instance in this pool is considered healthy if and only if the health checks pass. Only legacy HttpHealthChecks are supported. Only one health check may be specified. Defaults to None.
        backup_pool(str, Optional):
            The server-defined URL for the resource. This field is applicable only when the containing target pool is serving a forwarding rule as the primary pool, and its failoverRatio field is properly set to a value between [0, 1]. backupPool and failoverRatio together define the fallback behavior of the primary target pool: if the ratio of the healthy instances in the primary pool is at or below failoverRatio, traffic arriving at the load-balanced IP will be directed to the backup pool. In case where failoverRatio and backupPool are not set, or all the instances in the backup pool are unhealthy, the traffic will be directed back to the primary pool in the "force" mode, where traffic will be spread to the healthy instances with the best effort, or to all instances when no instance is healthy. Defaults to None.
        project(str, Optional):
            Project ID for this request.
        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              gcp.compute.target_pool.present:
                - name: value
                - region: value
                - project: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    # TODO uncomment below line, when implementation is added
    # project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    result["comment"].append(
        hub.tool.gcp.comment_utils.no_resource_create_update_comment(
            "gcp.compute.target_pool"
        )
    )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    project: str = None,
    region: str = None,
    request_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified target pool.

    Args:
        name (str): An Idem name of the resource.
        project(str, Optional):
            Project ID for this request.
        region (str, Optional):
            Name of the region scoping this request.
        request_id (str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.
        resource_id (str, optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              gcp.compute.target_pool.absent:
                - name: value
                - project: value
                - region: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    # TODO uncomment below line, when implementation is added
    # project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    result["comment"].append(
        hub.tool.gcp.comment_utils.no_resource_delete_comment("gcp.compute.target_pool")
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of target pools available to the specified project and region.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.target_pool
    """
    result = {}

    # TODO: Pagination
    describe_ret = await hub.exec.gcp.compute.target_pool.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.target_pool {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.target_pool.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""Metadata module for managing Crypto Key."""


def __init__(hub):
    hub.metadata.gcp.cloudkms.projects.locations.key_rings.crypto_keys.PATH = "projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}/cryptoKeys/{crypto_key_id}"



from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.generate.exec_param import ExecParam
from idem_gcp.tool.gcp.generate.scope import Scope


async def should_operate(hub, execution_context: ExecutionContext) -> bool:
    return True


async def operate(hub, execution_context: ExecutionContext) -> None:
    if execution_context.get_exec_param_value(ExecParam.SCOPED_FUNCTIONS) is not None:
        _find_scoped_method_name(execution_context)
    execution_context.method = hub.tool.gcp.resolver.resolve_sub(
        ".".join(
            ["gcp", execution_context.resource_type, execution_context.method_name]
        ),
        hub.tool.gcp.API,
    )


# modifies method_name to use appropriate one for scope
# removes redundant scope-related method parameters
def _find_scoped_method_name(execution_context: ExecutionContext):
    method_params = execution_context.method_params
    scope = Scope.GLOBAL
    if method_params.get("zone") is not None:
        scope = Scope.ZONAL
    else:
        # remove zone from method params, as it is not expected from non-zonal methods
        method_params.pop("zone", None)

    if method_params.get("region") is not None:
        scope = Scope.REGIONAL
    else:
        method_params.pop("region", None)

    scoped_functions = execution_context.get_exec_param_value(
        ExecParam.SCOPED_FUNCTIONS
    )

    if scoped_functions is not None and scoped_functions.get(scope) is not None:
        execution_context.method_name = scoped_functions.get(scope)



from typing import Any
from typing import Dict


class StateOperations:
    r"""Helper class for additional 'PATCH/UPDATE' operations that are executed during handling of present method
    to update properties the generic patch/methods don't handle
    """

    def __init__(
        self,
        hub,
        resource_type: str,
        patch_operations_dict: Dict[str, Any],
        result: Dict[str, Any],
        resource_body: Dict[str, Any],
    ):
        r"""
        Args:
            patch_operations_dict(Dict[str, Any]):
                A dictionary of additional operations to perform on the object identified by the key.
                The values are tuples with the first element - the method to call, the second element - arguments, third - the property is required in the end result
                fourth - should the property be removed from the request_body(some patch operations complain)
        """
        self.resource_type = resource_type
        self.patch_operations_dict = patch_operations_dict
        self.old_properties_dict = {
            k: result["old_state"].get(k)
            for k in patch_operations_dict.keys()
            if result["old_state"].get(k) is not None
        }
        self.changed_properties_dict = {
            k: {}
            if resource_body.get(k) is None
            else hub.tool.gcp.utils.compare_states(
                {k: self.old_properties_dict.get(k)},
                {k: resource_body.get(k)},
                resource_type,
            )
            for k in patch_operations_dict.keys()
        }

    async def run_operations(self) -> Dict[str, Any]:
        ret = None
        r"""run all relevant(updated) operations"""
        for k, op in self.patch_operations_dict.items():
            if self.changed_properties_dict[k]:
                ret = await op[0](*op[1])

                if not ret["result"]:
                    return {"result": False, "comment": ret["comment"]}

        return (
            {"result": True, "new_state": ret["ret"]}
            if ret and "ret" in ret
            else {"result": True}
        )

    def pre_process_resource_body(self, resource_body: Dict[str, Any]):
        r"""puts back the old properties in the resource body so that the API doesn't complain"""
        for k, v in self.patch_operations_dict.items():
            if 2 in v and v[2] and not resource_body.get(k):
                resource_body[k] = self.old_properties_dict[k]
            if 3 in v and v[3] and k in resource_body:
                resource_body.pop(k)



"""Metadata module for managing Disk Types."""

PATH = [
    "projects/{project}/zones/{zone}/diskTypes/{diskType}",
    "projects/{project}/regions/{region}/diskTypes/{diskType}",
]

NATIVE_RESOURCE_TYPE = [
    "compute.disk_types",
    "compute.region_disk_types",
]



from typing import Any
from typing import Dict
from typing import List


async def update_ip_cidr_range(
    hub,
    ctx,
    current_subnetwork: Dict[str, Any],
    new_ip_cidr_range: List[Dict[str, Any]],
) -> Dict[str, Any]:
    resource_id = current_subnetwork.get("resource_id")

    return await hub.exec.gcp.compute.subnetwork.update_ip_cidr_range(
        ctx,
        new_ip_cidr_range,
        current_subnetwork.get("project"),
        current_subnetwork.get("region"),
        current_subnetwork,
        resource_id,
    )



"""State module for managing Resource Policies."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import regional_absent

# prevent commit hook from removing the import
absent = regional_absent

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str = None,
    project: str = None,
    region: str = None,
    resource_id: str = None,
    description: str = None,
    snapshot_schedule_policy: make_dataclass(
        "ResourcePolicySnapshotSchedulePolicy",
        [
            (
                "schedule",
                make_dataclass(
                    "ResourcePolicySnapshotSchedulePolicySchedule",
                    [
                        (
                            "daily_schedule",
                            make_dataclass(
                                "ResourcePolicyDailyCycle",
                                [
                                    ("days_in_cycle", int, field(default=0)),
                                    ("start_time", str, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "weekly_schedule",
                            make_dataclass(
                                "ResourcePolicyWeeklyCycle",
                                [
                                    (
                                        "day_of_weeks",
                                        List[
                                            make_dataclass(
                                                "ResourcePolicyWeeklyCycleDayOfWeek",
                                                [
                                                    (
                                                        "start_time",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    ("day", str, field(default=None)),
                                                ],
                                            )
                                        ],
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "hourly_schedule",
                            make_dataclass(
                                "ResourcePolicyHourlyCycle",
                                [
                                    ("hours_in_cycle", int, field(default=0)),
                                    ("start_time", str, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "retention_policy",
                make_dataclass(
                    "ResourcePolicySnapshotSchedulePolicyRetentionPolicy",
                    [
                        ("max_retention_days", int, field(default=0)),
                        ("on_source_disk_delete", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "snapshotProperties",
                make_dataclass(
                    "ResourcePolicySnapshotSchedulePolicySnapshotProperties",
                    [
                        ("guest_flush", bool, field(default=False)),
                        ("chain_name", str, field(default=None)),
                        ("storage_locations", List[str], field(default=None)),
                        ("labels", Dict[str, str], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    group_placement_policy: make_dataclass(
        "ResourcePolicyGroupPlacementPolicy",
        [
            ("vm_count", int, field(default=0)),
            ("collocation", str, field(default=None)),
            ("availability_domain_count", int, field(default=0)),
        ],
    ) = None,
    instance_schedule_policy: make_dataclass(
        "ResourcePolicyInstanceSchedulePolicy",
        [
            ("start_time", str, field(default=None)),
            ("time_zone", str, field(default=None)),
            ("expiration_time", str, field(default=None)),
            (
                "vm_start_schedule",
                make_dataclass(
                    "ResourcePolicyInstanceSchedulePolicySchedule",
                    [
                        ("schedule", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "vm_stop_schedule",
                make_dataclass(
                    "ResourcePolicyInstanceSchedulePolicySchedule",
                    [
                        ("schedule", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates a new resource policy.

    You can use resource policies to schedule actions for some Compute Engine resources.
    For example, you can use them to schedule persistent disk snapshots.

    Args:
        name(str, Optional):
            An Idem name of the resource.

        project(str, Optional):
            Project ID for this request. Defaults to None.

        region(str, Optional):
            Name of the region for this request. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        description(str, Optional):
            An optional description of this resource. Defaults to None.

        snapshot_schedule_policy(Dict[str, Any], Optional):
            Resource policy for persistent disks for creating snapshots. Defaults to None.
            A snapshot schedule policy specifies when and how frequently snapshots are to be created for
            the target disk. Also specifies how many and how long these scheduled snapshots should be retained.

            * schedule(Dict[str, Any], Optional):
                A Vm Maintenance Policy specifies what kind of infrastructure maintenance we are allowed to perform on
                this VM and when. Schedule that is applied to disks covered by this policy.
                A schedule for disks where the scheduled operations are performed.

                * daily_schedule(Dict[str, Any], Optional):
                    Time window specified for daily operations.

                    * days_in_cycle(int, Optional):
                        Defines a schedule with units measured in days.
                        The value determines how many days pass between the start of each cycle.

                    * start_time(str, Optional):
                        Start time of the window. This must be in UTC format that resolves to one of
                        00:00, 04:00, 08:00, 12:00, 16:00, or 20:00. For example, both 13:00-5 and 08:00 are valid.

                * weekly_schedule(Dict[str, Any], Optional):
                    Time window specified for weekly operations.

                    * day_of_weeks(List[Dict[str, Any]], Optional):
                        Up to 7 intervals/windows, one for each day of the week.

                        * start_time(str, Optional):
                            Time within the window to start the operations.
                            It must be in format \"HH:MM\", where HH : [00-23] and MM : [00-00] GMT.

                        * day(str, Optional):
                            Defines a schedule that runs on specific days of the week. Specify one or more days.
                            The following options are available:
                            MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.

                * hourly_schedule(Dict[str, Any], Optional):
                    Time window specified for hourly operations.

                    * hours_in_cycle(int, Optional):
                        Defines a schedule with units measured in hours.
                        The value determines how many hours pass between the start of each cycle.

                    * start_time(str, Optional):
                        Time within the window to start the operations.
                        It must be in format \"HH:MM\", where HH : [00-23] and MM : [00-00] GMT.

            * retention_policy(Dict[str, Any], Optional):
                Retention policy applied to snapshots created by this resource policy.
                Policy for retention of scheduled snapshots.

                * max_retention_days(int, Optional):
                    Maximum age of the snapshot that is allowed to be kept.

                * on_source_disk_delete(str, Optional):
                    Specifies the behavior to apply to scheduled snapshots when the source disk is deleted.
                    The following options are available:
                    APPLY_RETENTION_POLICY, KEEP_AUTO_SNAPSHOTS, UNSPECIFIED_ON_SOURCE_DISK_DELETE


            * snapshot_properties(Dict[str, Any], Optional):
                Properties with which snapshots are created such as labels, encryption keys.
                Specified snapshot properties for scheduled snapshots created by this policy.

                * guest_flush(bool, Optional):
                    Indication to perform a 'guest aware' snapshot.

                * chain_name(str, Optional):
                    Chain name that the snapshot is created in.

                * storage_locations(List[str, str], Optional):
                    Cloud Storage bucket storage location of the auto snapshot (regional or multi-regional).

                * labels(List[str, str], Optional):
                    Labels to apply to scheduled snapshots. These can be later modified by the setLabels method.
                    Label values may be empty.

        group_placement_policy(Dict[str, Any], Optional):
            Resource policy for instances for placement configuration. Defaults to None.
            A GroupPlacementPolicy specifies resource placement configuration.
            It specifies the failure bucket separation as well as network locality

            * vm_count(int, Optional):
                Number of VMs in this placement group. Google does not recommend that you use this field unless
                you use a compact policy, and you want your policy to work only if it contains this exact number of VMs.

            * collocation(str, Optional):
                Specifies network collocation. The following options are available:
                COLLOCATED, UNSPECIFIED_COLLOCATION

            * availability_domain_count(int, Optional):
                The number of availability domains to spread instances across.
                If two instances are in different availability domain, they are not in the same low latency network.

        instance_schedule_policy(Dict[str, Any], Optional):
            Resource policy for scheduling instance operations. Defaults to None.
            An InstanceSchedulePolicy specifies when and how frequent certain operations are performed on the instance.

            * start_time(str, Optional):
                The start time of the schedule. The timestamp is an RFC3339 string.

            * time_zone(str, Optional):
                Specifies the time zone to be used in interpreting Schedule.schedule.
                The value of this field must be a time zone name from the tz database:
                http://en.wikipedia.org/wiki/Tz_database.

            * expiration_time(str, Optional):
                The expiration time of the schedule. The timestamp is an RFC3339 string.

            * vm_start_schedule(Dict[str, Any], Optional):
                Specifies the schedule for starting instances.

                * schedule(str, Optional):
                    Specifies the frequency for the operation, using the unix-cron format.

            * vm_stop_schedule(Dict[str, Any], Optional):
                Specifies the schedule for stopping instances.

                * schedule(str, Optional):
                    Specifies the frequency for the operation, using the unix-cron format.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_resource_name:
              gcp.compute.resource_policy.present:
                - name: value
                - project: value
                - region: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    # TODO uncomment below line, when implementation is added
    # project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    result["comment"].append(
        "No-op: There is no create/update function for gcp.compute.resource_policy"
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of resource policies contained within the specified region.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.resource_policy
    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.resource_policy.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe resource policies {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.resource_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""State module for managing HealthChecks."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

from idem_gcp.tool.gcp.utils import global_absent

# prevent commit hook from removing the import
absent = global_absent

__contracts__ = ["resource"]
RESOURCE_TYPE = "compute.health_check"
RESOURCE_TYPE_FULL = "gcp.compute.health_check"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    request_id: str = None,
    project: str = None,
    health_check: str = None,
    ssl_health_check: make_dataclass(
        "SSLHealthCheck",
        [
            ("port_name", str, field(default=None)),
            ("response", str, field(default=None)),
            ("proxy_header", str, field(default=None)),
            ("port_specification", str, field(default=None)),
            ("port", int, field(default=None)),
            ("request", str, field(default=None)),
        ],
    ) = None,
    https_health_check: make_dataclass(
        "HTTPSHealthCheck",
        [
            ("host", str, field(default=None)),
            ("port_specification", str, field(default=None)),
            ("port_name", str, field(default=None)),
            ("proxy_header", str, field(default=None)),
            ("request_path", str, field(default=None)),
            ("port", int, field(default=None)),
            ("response", str, field(default=None)),
        ],
    ) = None,
    type_: (str, "alias=type") = None,
    description: str = None,
    http2_health_check: make_dataclass(
        "HTTP2HealthCheck",
        [
            ("port_specification", str, field(default=None)),
            ("port", int, field(default=None)),
            ("port_name", str, field(default=None)),
            ("request_path", str, field(default=None)),
            ("proxy_header", str, field(default=None)),
            ("response", str, field(default=None)),
            ("host", str, field(default=None)),
        ],
    ) = None,
    tcp_health_check: make_dataclass(
        "TCPHealthCheck",
        [
            ("proxy_header", str, field(default=None)),
            ("port_specification", str, field(default=None)),
            ("response", str, field(default=None)),
            ("request", str, field(default=None)),
            ("port_name", str, field(default=None)),
            ("port", int, field(default=None)),
        ],
    ) = None,
    check_interval_sec: int = None,
    unhealthy_threshold: int = None,
    timeout_sec: int = None,
    log_config: make_dataclass(
        "HealthCheckLogConfig", [("enable", bool, field(default=None))]
    ) = None,
    grpc_health_check: make_dataclass(
        "GRPCHealthCheck",
        [
            ("port_specification", str, field(default=None)),
            ("grpc_service_name", str, field(default=None)),
            ("port_name", str, field(default=None)),
            ("port", int, field(default=None)),
        ],
    ) = None,
    http_health_check: make_dataclass(
        "HTTPHealthCheck",
        [
            ("request_path", str, field(default=None)),
            ("port", int, field(default=None)),
            ("host", str, field(default=None)),
            ("response", str, field(default=None)),
            ("port_specification", str, field(default=None)),
            ("port_name", str, field(default=None)),
            ("proxy_header", str, field(default=None)),
        ],
    ) = None,
    healthy_threshold: int = None,
) -> Dict[str, Any]:
    r"""Creates or updates a HealthCheck resource in the specified project using the data included in the request.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        project(str):
            Project ID for this request.

        ssl_health_check(Dict[str, Any], Optional):
            Defaults to None.

            * port_name (str, Optional):
                Not supported.

            * response (str, Optional):
                Creates a content-based SSL health check. In addition to establishing a TCP connection and the TLS handshake, you can configure the health check to pass only when the backend sends this exact response ASCII string, up to 1024 bytes in length. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#criteria-protocol-ssl-tcp

            * proxy_header (str, Optional):
                Specifies the type of proxy header to append before sending data to the backend, either NONE or PROXY_V1. The default is NONE.

            * port_specification (str, Optional):
                Specifies how a port is selected for health checking. Can be one of the following values: USE_FIXED_PORT: Specifies a port number explicitly using the port field in the health check. Supported by backend services for pass-through load balancers and backend services for proxy load balancers. Not supported by target pools. The health check supports all backends supported by the backend service provided the backend can be health checked. For example, GCE_VM_IP network endpoint groups, GCE_VM_IP_PORT network endpoint groups, and instance group backends. USE_NAMED_PORT: Not supported. USE_SERVING_PORT: Provides an indirect method of specifying the health check port by referring to the backend service. Only supported by backend services for proxy load balancers. Not supported by target pools. Not supported by backend services for pass-through load balancers. Supports all backends that can be health checked; for example, GCE_VM_IP_PORT network endpoint groups and instance group backends. For GCE_VM_IP_PORT network endpoint group backends, the health check uses the port number specified for each endpoint in the network endpoint group. For instance group backends, the health check uses the port number determined by looking up the backend service's named port in the instance group's list of named ports.
                    Enum type. Allowed values:
                        "USE_FIXED_PORT" - The port number in the health check's port is used for health checking. Applies to network endpoint group and instance group backends.
                        "USE_NAMED_PORT" - Not supported.
                        "USE_SERVING_PORT" - For network endpoint group backends, the health check uses the port number specified on each endpoint in the network endpoint group. For instance group backends, the health check uses the port number specified for the backend service's named port defined in the instance group's named ports.

            * port (int, Optional):
                The TCP port number to which the health check prober sends packets. The default value is 443. Valid values are 1 through 65535.

            * request (str, Optional):
                Instructs the health check prober to send this exact ASCII string, up to 1024 bytes in length, after establishing the TCP connection and SSL handshake.

        https_health_check(Dict[str, Any], Optional):
            Defaults to None.

            * host (str, Optional):
                The value of the host header in the HTTPS health check request. If left empty (default value), the host header is set to the destination IP address to which health check packets are sent. The destination IP address depends on the type of load balancer. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#hc-packet-dest

            * port_specification (str, Optional):
                Specifies how a port is selected for health checking. Can be one of the following values: USE_FIXED_PORT: Specifies a port number explicitly using the port field in the health check. Supported by backend services for pass-through load balancers and backend services for proxy load balancers. Not supported by target pools. The health check supports all backends supported by the backend service provided the backend can be health checked. For example, GCE_VM_IP network endpoint groups, GCE_VM_IP_PORT network endpoint groups, and instance group backends. USE_NAMED_PORT: Not supported. USE_SERVING_PORT: Provides an indirect method of specifying the health check port by referring to the backend service. Only supported by backend services for proxy load balancers. Not supported by target pools. Not supported by backend services for pass-through load balancers. Supports all backends that can be health checked; for example, GCE_VM_IP_PORT network endpoint groups and instance group backends. For GCE_VM_IP_PORT network endpoint group backends, the health check uses the port number specified for each endpoint in the network endpoint group. For instance group backends, the health check uses the port number determined by looking up the backend service's named port in the instance group's list of named ports.
                    Enum type. Allowed values:
                        "USE_FIXED_PORT" - The port number in the health check's port is used for health checking. Applies to network endpoint group and instance group backends.
                        "USE_NAMED_PORT" - Not supported.
                        "USE_SERVING_PORT" - For network endpoint group backends, the health check uses the port number specified on each endpoint in the network endpoint group. For instance group backends, the health check uses the port number specified for the backend service's named port defined in the instance group's named ports.

            * port_name (str, Optional):
                Not supported.

            * proxy_header (str, Optional):
                Specifies the type of proxy header to append before sending data to the backend, either NONE or PROXY_V1. The default is NONE.
                    Enum type. Allowed values:
                        "NONE"
                        "PROXY_V1"

            * request_path (str, Optional):
                The request path of the HTTPS health check request. The default value is /.

            * port (int, Optional):
                The TCP port number to which the health check prober sends packets. The default value is 443. Valid values are 1 through 65535.

            * response (str, Optional):
                Creates a content-based HTTPS health check. In addition to the required HTTP 200 (OK) status code, you can configure the health check to pass only when the backend sends this specific ASCII response string within the first 1024 bytes of the HTTP response body. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#criteria-protocol-http

        type(str, Optional):
            Specifies the type of the healthCheck, either TCP, SSL, HTTP, HTTPS, HTTP2 or GRPC. Exactly one of the protocol-specific health check fields must be specified, which must match type field.
                Enum type. Allowed values:
                    "GRPC"
                    "HTTP"
                    "HTTP2"
                    "HTTPS"
                    "INVALID"
                    "SSL"
                    "TCP". Defaults to None.

        description(str, Optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.

        http2_health_check(Dict[str, Any], Optional):
            Defaults to None.

            * port_specification (str, Optional):
                Specifies how a port is selected for health checking. Can be one of the following values: USE_FIXED_PORT: Specifies a port number explicitly using the port field in the health check. Supported by backend services for pass-through load balancers and backend services for proxy load balancers. Not supported by target pools. The health check supports all backends supported by the backend service provided the backend can be health checked. For example, GCE_VM_IP network endpoint groups, GCE_VM_IP_PORT network endpoint groups, and instance group backends. USE_NAMED_PORT: Not supported. USE_SERVING_PORT: Provides an indirect method of specifying the health check port by referring to the backend service. Only supported by backend services for proxy load balancers. Not supported by target pools. Not supported by backend services for pass-through load balancers. Supports all backends that can be health checked; for example, GCE_VM_IP_PORT network endpoint groups and instance group backends. For GCE_VM_IP_PORT network endpoint group backends, the health check uses the port number specified for each endpoint in the network endpoint group. For instance group backends, the health check uses the port number determined by looking up the backend service's named port in the instance group's list of named ports.
                    Enum type. Allowed values:
                        "USE_FIXED_PORT" - The port number in the health check's port is used for health checking. Applies to network endpoint group and instance group backends.
                        "USE_NAMED_PORT" - Not supported.
                        "USE_SERVING_PORT" - For network endpoint group backends, the health check uses the port number specified on each endpoint in the network endpoint group. For instance group backends, the health check uses the port number specified for the backend service's named port defined in the instance group's named ports.

            * port (int, Optional):
                The TCP port number to which the health check prober sends packets. The default value is 443. Valid values are 1 through 65535.

            * port_name (str, Optional):
                Not supported.

            * request_path (str, Optional):
                The request path of the HTTP/2 health check request. The default value is /.

            * proxy_header (str, Optional):
                Specifies the type of proxy header to append before sending data to the backend, either NONE or PROXY_V1. The default is NONE.
                    Enum type. Allowed values:
                        "NONE"
                        "PROXY_V1"

            * response (str, Optional):
                Creates a content-based HTTP/2 health check. In addition to the required HTTP 200 (OK) status code, you can configure the health check to pass only when the backend sends this specific ASCII response string within the first 1024 bytes of the HTTP response body. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#criteria-protocol-http

            * host (str, Optional):
                The value of the host header in the HTTP/2 health check request. If left empty (default value), the host header is set to the destination IP address to which health check packets are sent. The destination IP address depends on the type of load balancer. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#hc-packet-dest

        tcp_health_check(Dict[str, Any], Optional):
            Defaults to None.

            * proxy_header (str, Optional):
                Specifies the type of proxy header to append before sending data to the backend, either NONE or PROXY_V1. The default is NONE.
                    Enum type. Allowed values:
                        "NONE"
                        "PROXY_V1"

            * port_specification (str, Optional):
                Specifies how a port is selected for health checking. Can be one of the following values: USE_FIXED_PORT: Specifies a port number explicitly using the port field in the health check. Supported by backend services for pass-through load balancers and backend services for proxy load balancers. Not supported by target pools. The health check supports all backends supported by the backend service provided the backend can be health checked. For example, GCE_VM_IP network endpoint groups, GCE_VM_IP_PORT network endpoint groups, and instance group backends. USE_NAMED_PORT: Not supported. USE_SERVING_PORT: Provides an indirect method of specifying the health check port by referring to the backend service. Only supported by backend services for proxy load balancers. Not supported by target pools. Not supported by backend services for pass-through load balancers. Supports all backends that can be health checked; for example, GCE_VM_IP_PORT network endpoint groups and instance group backends. For GCE_VM_IP_PORT network endpoint group backends, the health check uses the port number specified for each endpoint in the network endpoint group. For instance group backends, the health check uses the port number determined by looking up the backend service's named port in the instance group's list of named ports.
                    Enum type. Allowed values:
                        "USE_FIXED_PORT" - The port number in the health check's port is used for health checking. Applies to network endpoint group and instance group backends.
                        "USE_NAMED_PORT" - Not supported.
                        "USE_SERVING_PORT" - For network endpoint group backends, the health check uses the port number specified on each endpoint in the network endpoint group. For instance group backends, the health check uses the port number specified for the backend service's named port defined in the instance group's named ports.
            * response (str, Optional):
                Creates a content-based TCP health check. In addition to establishing a TCP connection, you can configure the health check to pass only when the backend sends this exact response ASCII string, up to 1024 bytes in length. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#criteria-protocol-ssl-tcp
            * request (str, Optional):
                Instructs the health check prober to send this exact ASCII string, up to 1024 bytes in length, after establishing the TCP connection.
            * port_name (str, Optional):
                Not supported.
            * port (int, Optional):
                The TCP port number to which the health check prober sends packets. The default value is 80. Valid values are 1 through 65535.

        check_interval_sec(int, Optional):
            How often (in seconds) to send a health check. The default value is 5 seconds. Defaults to None.

        unhealthy_threshold(int, Optional):
            A so-far healthy instance will be marked unhealthy after this many consecutive failures. The default value is 2. Defaults to None.

        timeout_sec(int, Optional):
            How long (in seconds) to wait before claiming failure. The default value is 5 seconds. It is invalid for timeoutSec to have greater value than checkIntervalSec. Defaults to None.

        log_config(Dict[str, Any], Optional):
            Configure logging on this health check.

            HealthCheckLogConfig: Configuration of logging on a health check. If logging is enabled, logs will be exported to Stackdriver. Defaults to None.
            * enable (bool, Optional):
                Indicates whether or not to export logs. This is false by default, which means no health check logging will be done.

        grpc_health_check(Dict[str, Any], Optional):
            Defaults to None.

            * port_specification (str, Optional):
                Specifies how a port is selected for health checking. Can be one of the following values: USE_FIXED_PORT: Specifies a port number explicitly using the port field in the health check. Supported by backend services for pass-through load balancers and backend services for proxy load balancers. Not supported by target pools. The health check supports all backends supported by the backend service provided the backend can be health checked. For example, GCE_VM_IP network endpoint groups, GCE_VM_IP_PORT network endpoint groups, and instance group backends. USE_NAMED_PORT: Not supported. USE_SERVING_PORT: Provides an indirect method of specifying the health check port by referring to the backend service. Only supported by backend services for proxy load balancers. Not supported by target pools. Not supported by backend services for pass-through load balancers. Supports all backends that can be health checked; for example, GCE_VM_IP_PORT network endpoint groups and instance group backends. For GCE_VM_IP_PORT network endpoint group backends, the health check uses the port number specified for each endpoint in the network endpoint group. For instance group backends, the health check uses the port number determined by looking up the backend service's named port in the instance group's list of named ports.
                    Enum type. Allowed values:
                        "USE_FIXED_PORT" - The port number in the health check's port is used for health checking. Applies to network endpoint group and instance group backends.
                        "USE_NAMED_PORT" - Not supported.
                        "USE_SERVING_PORT" - For network endpoint group backends, the health check uses the port number specified on each endpoint in the network endpoint group. For instance group backends, the health check uses the port number specified for the backend service's named port defined in the instance group's named ports.

            * grpc_service_name (str, Optional):
                The gRPC service name for the health check. This field is optional. The value of grpc_service_name has the following meanings by convention: - Empty service_name means the overall status of all services at the backend. - Non-empty service_name means the health of that gRPC service, as defined by the owner of the service. The grpc_service_name can only be ASCII.

            * port_name (str, Optional):
                Not supported.

            * port (int, Optional):
                The TCP port number to which the health check prober sends packets. Valid values are 1 through 65535.

        http_health_check(Dict[str, Any], Optional):
            Defaults to None.

            * request_path (str, Optional):
                The request path of the HTTP health check request. The default value is /.

            * port (int, Optional):
                The TCP port number to which the health check prober sends packets. The default value is 80. Valid values are 1 through 65535.

            * host (str, Optional):
                The value of the host header in the HTTP health check request. If left empty (default value), the host header is set to the destination IP address to which health check packets are sent. The destination IP address depends on the type of load balancer. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#hc-packet-dest

            * response (str, Optional):
                Creates a content-based HTTP health check. In addition to the required HTTP 200 (OK) status code, you can configure the health check to pass only when the backend sends this specific ASCII response string within the first 1024 bytes of the HTTP response body. For details, see: https://cloud.google.com/load-balancing/docs/health-check-concepts#criteria-protocol-http

            * port_specification (str, Optional):
                Specifies how a port is selected for health checking. Can be one of the following values: USE_FIXED_PORT: Specifies a port number explicitly using the port field in the health check. Supported by backend services for pass-through load balancers and backend services for proxy load balancers. Also supported in legacy HTTP health checks for target pools. The health check supports all backends supported by the backend service provided the backend can be health checked. For example, GCE_VM_IP network endpoint groups, GCE_VM_IP_PORT network endpoint groups, and instance group backends. USE_NAMED_PORT: Not supported. USE_SERVING_PORT: Provides an indirect method of specifying the health check port by referring to the backend service. Only supported by backend services for proxy load balancers. Not supported by target pools. Not supported by backend services for pass-through load balancers. Supports all backends that can be health checked; for example, GCE_VM_IP_PORT network endpoint groups and instance group backends. For GCE_VM_IP_PORT network endpoint group backends, the health check uses the port number specified for each endpoint in the network endpoint group. For instance group backends, the health check uses the port number determined by looking up the backend service's named port in the instance group's list of named ports.
                    Enum type. Allowed values:
                        "USE_FIXED_PORT" - The port number in the health check's port is used for health checking. Applies to network endpoint group and instance group backends.
                        "USE_NAMED_PORT" - Not supported.
                        "USE_SERVING_PORT" - For network endpoint group backends, the health check uses the port number specified on each endpoint in the network endpoint group. For instance group backends, the health check uses the port number specified for the backend service's named port defined in the instance group's named ports.

            * port_name (str, Optional):
                Not supported.

            * proxy_header (str, Optional):
                Specifies the type of proxy header to append before sending data to the backend, either NONE or PROXY_V1. The default is NONE.
                    Enum type. Allowed values:
                        "NONE"
                        "PROXY_V1"

        healthy_threshold(int, Optional):
            A so-far unhealthy instance will be marked healthy after this many consecutive successes. The default value is 2. Defaults to None.

        health_check(str):
            Name of the HealthCheck resource to return.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              gcp.compute.health_check.present:
                - name: value
                - project: value
                - type_: TCP
                - tcp_health_check:
                    port: 80
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
    resource_type_camel = hub.tool.gcp.case.camel(RESOURCE_TYPE_FULL.split(".")[-1])

    if hub.tool.gcp.resource_prop_utils.properties_mismatch_resource_id(
        RESOURCE_TYPE, resource_id, {**locals(), resource_type_camel: name}
    ):
        result["comment"].append(
            hub.tool.gcp.comment_utils.properties_mismatch_resource_id_comment(
                RESOURCE_TYPE_FULL, name
            )
        )

    # Handle operation(s) in progress, if any
    if ctx.get("rerun_data"):
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), RESOURCE_TYPE
        )

        if not handle_operation_ret["result"]:
            result["comment"] += handle_operation_ret["comment"]
            if handle_operation_ret.get("rerun_data"):
                result["rerun_data"] = handle_operation_ret["rerun_data"]
                if handle_operation_ret["rerun_data"].get("has_error", False):
                    result["result"] = False
            else:
                result["result"] = False

            return result

        resource_id = handle_operation_ret["resource_id"]

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )

    if resource_id:
        old_get_ret = await hub.exec.gcp.compute.health_check.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"] or (
            not old_get_ret["ret"]
            and (ctx["rerun_data"] or get_resource_only_with_resource_id)
        ):
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        # long running operation has succeeded - both update and create
        if ctx.get("rerun_data"):
            result["new_state"] = old_get_ret["ret"]
            result["old_state"] = ctx["rerun_data"]["old_state"]
            if result["old_state"]:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.update_comment(RESOURCE_TYPE_FULL, name)
                )
            else:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.create_comment(RESOURCE_TYPE_FULL, name)
                )
            return result

        result["old_state"] = old_get_ret["ret"]
    elif not get_resource_only_with_resource_id:
        resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
            RESOURCE_TYPE, {**locals(), "healthCheck": name}
        )
        get_ret = await hub.exec.gcp.compute.health_check.get(
            ctx, resource_id=resource_id
        )

        if not get_ret["result"]:
            result["result"] = False
            result["comment"] += get_ret["comment"]
            return result

        if get_ret["ret"]:
            result["old_state"] = get_ret["ret"]

    resource_body = {
        "log_config": log_config,
        "unhealthy_threshold": unhealthy_threshold,
        "name": name,
        "timeout_sec": timeout_sec,
        "description": description,
        "http_health_check": http_health_check,
        "http2_health_check": http2_health_check,
        "grpc_health_check": grpc_health_check,
        "tcp_health_check": tcp_health_check,
        "check_interval_sec": check_interval_sec,
        "healthy_threshold": healthy_threshold,
        "https_health_check": https_health_check,
        "ssl_health_check": ssl_health_check,
        "type_": type_,
    }

    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}
    operation = None
    if result["old_state"]:
        resource_body["resource_id"] = resource_id
        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            resource_body,
            RESOURCE_TYPE,
        )

        if not changes:
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(RESOURCE_TYPE_FULL, name)
            )
            result["new_state"] = result["old_state"]
            return result

        changed_non_updatable_properties = (
            hub.tool.gcp.resource_prop_utils.get_changed_non_updatable_properties(
                RESOURCE_TYPE, changes
            )
        )
        if changed_non_updatable_properties:
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                    RESOURCE_TYPE_FULL,
                    name,
                    changed_non_updatable_properties,
                )
            )
            result["new_state"] = result["old_state"]
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    RESOURCE_TYPE_FULL, name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            return result

        # Perform update
        update_ret = await hub.exec.gcp_api.client.compute.health_check.update(
            hub,
            ctx,
            name=name,
            resource_id=resource_id,
            request_id=request_id,
            body=resource_body,
        )
        if not update_ret["result"] or not update_ret["ret"]:
            result["result"] = False
            result["comment"] += update_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(update_ret["ret"]):
            operation = update_ret["ret"]
    else:
        if ctx.get("test", False):
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    RESOURCE_TYPE_FULL, name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            result["new_state"]["resource_id"] = resource_id
            return result

        # Create
        create_ret = await hub.exec.gcp_api.client.compute.health_check.insert(
            ctx,
            name=name,
            project=project,
            request_id=request_id,
            body=resource_body,
        )
        if not create_ret["result"] or not create_ret["ret"]:
            result["result"] = False
            if create_ret["comment"] and next(
                (
                    comment
                    for comment in create_ret["comment"]
                    if "alreadyExists" in comment
                ),
                None,
            ):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        RESOURCE_TYPE_FULL, name
                    )
                )
            else:
                result["comment"] += create_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(create_ret["ret"]):
            operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.global_operation"
        )
        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves the list of HealthCheck resources available to the specified project.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.health_check
    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.health_check.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe {RESOURCE_TYPE_FULL} {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")

        result[resource_id] = {
            f"{RESOURCE_TYPE_FULL}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



from typing import Dict

from idem_gcp.tool.gcp.schema.results_collector import ResultsCollector


def check(hub, current_schema: Dict) -> ResultsCollector:
    collector = ResultsCollector()

    for method in current_schema.keys():
        method_name = method.split(".")[-1]
        if method_name in {"field", "make_dataclass"}:
            continue
        method_path = f"root['{method}']"
        if not hub.tool.gcp.case.is_snake_case(method):
            collector.add_breaking(
                schema_path=method_path, description="method name is not in snake_case"
            )
        parameters: Dict = current_schema.get(method).get("parameters") or {}
        for parameter in parameters.keys():
            parameter_path = f"{method_path}['parameters']['{parameter}']"
            if not hub.tool.gcp.case.is_snake_case(parameter):
                collector.add_breaking(
                    schema_path=parameter_path,
                    description="parameter name is not in snake_case",
                )
            if not hub.tool.gcp.case.is_unclashed(parameter):
                collector.add_breaking(
                    schema_path=parameter_path,
                    description="parameter name shadows builtin",
                )

    return collector



"""Exec module for managing Subnetworks."""
__func_alias__ = {"list_": "list"}

from typing import Any, Dict, List

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext


async def list_(
    hub,
    ctx,
    project: str = None,
    region: str = None,
    filter_: (str, "alias=filter") = None,
    order_by: str = None,
):
    r"""Retrieves a list of subnetworks available to the specified project.

    Args:
        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            Name of the region scoping this request.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name.

    Examples:
        random-name:
          exec.run:
          - path: gcp.compute.subnetwork.list
          - kwargs:
              project: project-name
              region: region-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if region:
        list_result = await hub.exec.gcp_api.client.compute.subnetwork.list(
            ctx,
            project=project,
            region=region,
            filter=filter_,
            orderBy=order_by,
        )
    else:
        list_result = await hub.exec.gcp_api.client.compute.subnetwork.aggregatedList(
            ctx,
            project=project,
            filter=filter_,
            orderBy=order_by,
        )

    result["comment"] += list_result["comment"]
    if not list_result["result"]:
        result["result"] = False
        return result

    result["ret"] = list_result["ret"].get("items", [])
    return result


async def get(
    hub,
    ctx,
    project: str = None,
    region: str = None,
    name: str = None,
    resource_id: str = None,
):
    r"""Returns the specified subnetwork. Gets a list of available subnetworks list() request.

    Args:
        project(str, Optional):
            Project ID for this request. Defaults to None.

        region(str, Optional):
            Name of the region scoping this request. Defaults to None.

        name(str, Optional):
            Name of the Subnetwork resource to return. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Examples:
        random-name:
          exec.run:
          - path: gcp.compute.subnetwork.get
          - kwargs:
              project: project-name
              region: region-name
              name: subnetwork-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if resource_id:
        ret = await hub.exec.gcp_api.client.compute.subnetwork.get(
            ctx,
            resource_id=resource_id,
        )
    elif project and region and name:
        ret = await hub.exec.gcp_api.client.compute.subnetwork.get(
            ctx, project=project, region=region, subnetwork=name, name=name
        )
    else:
        result["result"] = False
        result["comment"] = [
            f"gcp.compute.subnetwork {name} either resource_id or project, region and name"
            f" should be specified."
        ]
        return result

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def update_ip_cidr_range(
    hub,
    ctx,
    new_ip_cidr_range: List[Dict[str, Any]],
    project: str = None,
    region: str = None,
    subnetwork: str = None,
    resource_id: str = None,
    request_id: str = None,
):
    r"""Expands the IP CIDR range of the subnetwork to a specified value.

    Args:
        new_ip_cidr_range: The IP (in CIDR format or netmask) of internal addresses that are legal on
        this Subnetwork. This range should be disjoint from other subnetworks within this network.
        This range can only be larger than (i.e. a superset of) the range previously defined before the update.

        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            Name of the region scoping this request.

        subnetwork(str, Optional):
            Name of the Subnetwork resource to update.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

        request_id(str, Optional): An optional request ID to identify requests. Specify a unique request ID so that
        if you must retry your request, the server will know to ignore the request if it has already been completed.
        For example, consider a situation where you make an initial request and the request times out. If you make
        the request again with the same request ID, the server can check if original operation with the same request
        ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating
        duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    """
    request_body = {"ip_cidr_range": new_ip_cidr_range}
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
    execution_context = ExecutionContext(
        resource_type="compute.subnetwork",
        method_name="expandIpCidrRange",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "request_id": request_id,
            "project": project,
            "region": region,
            "subnetwork": subnetwork,
            "body": request_body,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)



from typing import Any
from typing import Dict
from typing import List


async def update_size_gb(
    hub,
    ctx,
    current_size_gb: str,
    desired_size_gb: str,
    resource_id: str,
    request_id: str,
) -> Dict[str, Any]:
    result = {"result": True, "comment": []}

    if desired_size_gb is not None and str(current_size_gb) != str(desired_size_gb):
        if not ctx.get("test"):
            resize_body = {"size_gb": desired_size_gb}
            update_ret = await hub.exec.gcp_api.client.compute.disk.resize(
                ctx,
                resource_id=resource_id,
                body=resize_body,
                request_id=request_id,
            )

            if not update_ret["result"]:
                result["result"] = False
                result["comment"] += update_ret["comment"]
                return result

            if hub.tool.gcp.operation_utils.is_operation(update_ret["ret"]):
                operation = update_ret["ret"]
                operation_type = hub.tool.gcp.operation_utils.get_operation_type(
                    operation.get("selfLink")
                )
                operation_id = (
                    hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                        operation.get("selfLink"), operation_type
                    )
                )
                handle_operation_ret = (
                    await hub.tool.gcp.operation_utils.handle_operation(
                        ctx,
                        {"operation_id": operation_id},
                        "compute.disk",
                        True,
                    )
                )
                if not handle_operation_ret["result"]:
                    result["result"] = False
                    result["comment"] += handle_operation_ret["comment"]
                    return result

    return result


async def update_labels(
    hub,
    ctx,
    current_labels: Dict[str, str],
    new_labels: Dict[str, str],
    label_fingerprint: str,
    request_id: str,
    project: str,
    name: str,
    zone: str,
    region: str,
) -> Dict[str, Any]:
    result = {"result": True, "comment": []}

    if (
        new_labels is not None
        and not (new_labels == {} and current_labels is None)
        and current_labels != new_labels
    ):
        if not ctx.get("test"):
            set_labels_body = {
                "labels": new_labels,
                "label_fingerprint": label_fingerprint,
            }
            kwargs = {
                "project": project,
                "resource": name,
                "body": set_labels_body,
                "request_id": request_id,
            }
            if zone:
                kwargs.update({"zone": zone})
            if region:
                kwargs.update({"region": region})

            update_ret = await hub.exec.gcp_api.client.compute.disk.setLabels(
                ctx,
                **kwargs,
            )

            if not update_ret["result"]:
                result["result"] = False
                result["comment"] += update_ret["comment"]
                return result

            if hub.tool.gcp.operation_utils.is_operation(update_ret["ret"]):
                operation = update_ret["ret"]
                operation_type = hub.tool.gcp.operation_utils.get_operation_type(
                    operation.get("selfLink")
                )
                operation_id = (
                    hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                        operation.get("selfLink"), operation_type
                    )
                )
                handle_operation_ret = (
                    await hub.tool.gcp.operation_utils.handle_operation(
                        ctx,
                        {"operation_id": operation_id},
                        "compute.disk",
                        True,
                    )
                )
                if not handle_operation_ret["result"]:
                    result["result"] = False
                    result["comment"] += handle_operation_ret["comment"]
                    return result

    return result


async def update_resource_policies(
    hub,
    ctx,
    current_resource_policies: List[str],
    desired_resource_policies: List[str],
    resource_id: str,
    request_id: str,
) -> Dict[str, Any]:
    result = {"result": True, "comment": []}

    if not hub.tool.gcp.state_comparison_utils.are_lists_identical(
        current_resource_policies, desired_resource_policies
    ):
        # If (some) resource policies are links, parse them to resource policy resource ids
        new_resource_policies = [
            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                policy, "compute.resource_policy"
            )
            for policy in (
                desired_resource_policies
                if desired_resource_policies is not None
                else []
            )
        ]
        current_resource_policies = [
            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                policy, "compute.resource_policy"
            )
            for policy in current_resource_policies or []
        ]

        resource_policies_to_remove = [
            policy
            for policy in current_resource_policies
            if policy not in new_resource_policies
        ]
        resource_policies_to_add = [
            policy
            for policy in new_resource_policies
            if policy not in current_resource_policies
        ]

        if len(resource_policies_to_remove) > 0 or len(resource_policies_to_add) > 0:
            if not ctx.get("test"):
                add_request_body = None
                if len(resource_policies_to_add) > 0:
                    add_request_body = {"resource_policies": resource_policies_to_add}
                remove_request_body = None
                if len(resource_policies_to_remove) > 0:
                    remove_request_body = {
                        "resource_policies": resource_policies_to_remove
                    }

                remove_ret = None
                if remove_request_body is not None:
                    remove_ret = await hub.exec.gcp_api.client.compute.disk.removeResourcePolicies(
                        ctx,
                        resource_id=resource_id,
                        body=remove_request_body,
                        request_id=request_id,
                    )

                    if not remove_ret["result"]:
                        result["result"] = False
                        result["comment"] += remove_ret["comment"]
                        return result

                    if hub.tool.gcp.operation_utils.is_operation(remove_ret["ret"]):
                        operation = remove_ret["ret"]
                        operation_type = (
                            hub.tool.gcp.operation_utils.get_operation_type(
                                operation.get("selfLink")
                            )
                        )
                        operation_id = (
                            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                                operation.get("selfLink"), operation_type
                            )
                        )
                        handle_operation_ret = (
                            await hub.tool.gcp.operation_utils.handle_operation(
                                ctx,
                                {"operation_id": operation_id},
                                "compute.disk",
                                True,
                            )
                        )
                        if not handle_operation_ret["result"]:
                            result["result"] = False
                            result["comment"] += handle_operation_ret["comment"]
                            return result

                if (
                    remove_ret is None or remove_ret["result"]
                ) and add_request_body is not None:
                    add_ret = (
                        await hub.exec.gcp_api.client.compute.disk.addResourcePolicies(
                            ctx,
                            resource_id=resource_id,
                            body=add_request_body,
                            request_id=request_id,
                        )
                    )

                    if not add_ret["result"]:
                        result["result"] = False
                        result["comment"] += add_ret["comment"]
                        return result

                    if hub.tool.gcp.operation_utils.is_operation(add_ret["ret"]):
                        operation = add_ret["ret"]
                        operation_type = (
                            hub.tool.gcp.operation_utils.get_operation_type(
                                operation.get("selfLink")
                            )
                        )
                        operation_id = (
                            hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                                operation.get("selfLink"), operation_type
                            )
                        )
                        handle_operation_ret = (
                            await hub.tool.gcp.operation_utils.handle_operation(
                                ctx,
                                {"operation_id": operation_id},
                                "compute.disk",
                                True,
                            )
                        )
                        if not handle_operation_ret["result"]:
                            result["comment"] += handle_operation_ret["comment"]
                            return result

    return result



from inspect import signature

from pop.loader import LoadedMod

_HUB_SUB_CACHE = {}


def resolve(hub, ref: str):
    def gen_sub():
        call_path = ref.split(".")
        method_name = call_path[-1]
        hub_api_sub = hub
        # TODO: The first segment of the call path is ignored (assuming it's hub.** call).
        for c in call_path[1:]:
            hub_api_sub = getattr(hub_api_sub, c)

        _HUB_SUB_CACHE[ref] = HubContractMod(hub, hub_api_sub, method_name, ref)
        return _HUB_SUB_CACHE[ref]

    return _HUB_SUB_CACHE.get(ref) or gen_sub()


class HubContractMod(LoadedMod):
    def __init__(self, hub, hub_method, method_name, ref):
        super().__init__(name=ref)
        self._hub = hub
        self._hub_method = hub_method
        self._method_name = method_name

    @property
    def signature(self):
        return signature(self.__call__)

    def _missing(self, item: str):
        return self

    async def __call__(self, ctx, *args, **kwargs):
        result = {"result": True, "ret": None, "comment": []}

        try:
            response = await self._hub_method(ctx, *args, **kwargs)
            if not response:
                result["result"] = False
                return result

            result["result"] = response["result"]
            result["ret"] = response["ret"]
            result["comment"] += response["comment"]

        except Exception as e:
            result["result"] = False
            result["comment"].append(str(e))

        return result



"""Metadata module for managing Crypto Key Versions."""

PATH = "projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}/cryptoKeys/{crypto_key_id}/cryptoKeyVersions/{crypto_key_version_id}"



from typing import Any
from typing import Dict

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext


def __init__(hub):
    hub.tool.gcp.generate.PRE_EXECS_ORDER = [
        "resource_id_extractor",
        "parameter_sanitizer",
        "function_getter",
    ]
    hub.tool.gcp.generate.POST_EXECS_ORDER = ["operation_processor", "item_getter"]


async def execute(hub, execution_context: ExecutionContext) -> Dict[str, Any]:
    for operator_name in hub.tool.gcp.generate.PRE_EXECS_ORDER:
        operator = hub.tool.gcp.generate.generic_exec.get_pre_exec_operator(
            operator_name
        )
        if await operator.should_operate(execution_context):
            try:
                await operator.operate(execution_context)
            except ValueError as e:
                return {"result": False, "ret": None, "comment": [str(e)]}

    execution_context.response = await execution_context.method(
        **execution_context.method_params
    )

    for operator_name in hub.tool.gcp.generate.POST_EXECS_ORDER:
        operator = hub.tool.gcp.generate.generic_exec.get_post_exec_operator(
            operator_name
        )
        if await operator.should_operate(execution_context):
            await operator.operate(execution_context)

    return execution_context.response


def get_pre_exec_operator(hub, name: str):
    return getattr(hub.tool.gcp.generate.operators.pre_exec, name)


def get_post_exec_operator(hub, name: str):
    return getattr(hub.tool.gcp.generate.operators.post_exec, name)



from typing import Any
from typing import Dict

import aiohttp


async def request(
    hub, ctx, method: str, path: str, query_params: Dict[str, str], data: Dict[str, Any]
):
    url = "/".join((ctx.acct.endpoint_url, path))
    async with aiohttp.ClientSession(
        loop=hub.pop.Loop, raise_for_status=True, auth=ctx.acct.auth
    ) as session:
        async with session.request(
            url=url,
            method=method.lower(),
            allow_redirects=True,
            params=query_params,
            data=data,
        ) as response:
            return await response.json()



"""Metadata module for managing Crypto Key."""


def __init__(hub):
    hub.metadata.gcp.cloudkms.projects.locations.key_rings.import_jobs.PATH = "projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}/importJobs/{import_job_id}"



"""Metadata module for managing Disks."""

PATH = [
    "projects/{project}/zones/{zone}/disks/{disk}",
    "projects/{project}/regions/{region}/disks/{disk}",
]

NATIVE_RESOURCE_TYPE = [
    "compute.disks",
    "compute.regionDisks",
]



LIST_RESOURCES_WITH_PRESENT_WRAPPER = [
    "gcp.compute.instance",
    "gcp.compute.disk",
    "gcp.compute.forwarding_rule",
    "gcp.compute.subnetwork",
    "gcp.compute.backend_service",
    "gcp.compute.machine_image",
    "gcp.compute.instance_group",
    "gcp.compute.snapshot",
]

LIST_RESOURCES_WITH_ABSENT_WRAPPER = [
    "gcp.compute.disk",
    "gcp.compute.firewall",
    "gcp.compute.forwarding_rule",
    "gcp.compute.health_check",
    "gcp.compute.image",
    "gcp.compute.instance",
    "gcp.compute.instance_group",
    "gcp.compute.machine_image",
    "gcp.compute.network",
    "gcp.compute.node_template",
    "gcp.compute.reservation",
    "gcp.compute.resource_policy",
    "gcp.compute.snapshot",
    "gcp.compute.subnetwork",
]


async def call_present(hub, ctx):
    r"""Wrapper for present function.

    This method handles the parameters given in the ctx and calls the real present method only once when there is no
    rerun_data in the context and the resource needs to be created or updated, otherwise the method just reconciles
    until the particular operation created when present method was called is done. Once the operation is done,
    the resource_id is taken from the operation resource target link.

    First validation that is done inside the method is the verification whether the resource_id matches the
    properties given in the sls. If resource_id do NOT match, a warning comment is appended in order to notify the
    user about the difference. If resource_id matches it just continue the validation.

    Second validation is based on the presence of the resource_id. If resource_id is NOT in the ctx, it is going to
    be constructed using the parameter given in the SLS only if get_resource_only_with_resource_id flag is NOT set.

    If resource_id is NOT provided in the ctx and it cannot be constructed,
    it means that the resource does NOT exist in the Cloud, so call_present calls the real present method
    without result["old_state"], assuming that the user wants to create the resource.

    If resource_id is provided or constructed, then it is assumed that the resource already exists in the Cloud.
    Then inside the function, we GET the very last state of the resource using the resource_id and
    call present method assuming that the user wants to update an already existing resource.

    Args:
        hub: The redistributed pop central hub. The root of the namespace that pop operates on.
        ctx: Invocation context for this command.


    Returns:
        The result of a state with an appropriate comment for resource creation or update.
        If there is an error in the rerun_date then we return result["resutl"]=False and result["comment"]
        with the error message.

    """
    name = ctx.kwargs.get("name", None)
    state_ctx = ctx.kwargs.get("ctx")
    assert state_ctx, f"state context is missing: {state_ctx}"

    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    gcp_service_resource_type = state_ctx.get("tag").split("_|")[0]
    # TODO: This needs to be removed once all resources follow the contract
    if gcp_service_resource_type not in LIST_RESOURCES_WITH_PRESENT_WRAPPER:
        return await ctx.func(*ctx.args, **ctx.kwargs)

    if not ctx.kwargs.get("project"):
        ctx.kwargs["project"] = hub.tool.gcp.utils.get_project_from_account(
            state_ctx, ctx.kwargs.get("project")
        )

    service_resource_type = gcp_service_resource_type.replace("gcp.", "")
    resource_type_camel = hub.tool.gcp.case.camel(
        gcp_service_resource_type.split(".")[-1]
    )

    resource_path = service_resource_type.split(".")
    hub_ref_exec = hub.exec.gcp
    for resource_path_segment in resource_path:
        hub_ref_exec = hub_ref_exec[resource_path_segment]

    resource_id = (
        (ctx.kwargs.get("resource_id") or {})
        or (state_ctx.get("old_state") or {}).get("resource_id")
        or (state_ctx.get("rerun_data") or {}).get("resource_id")
    )
    local_params = {**ctx.kwargs}
    local_params.update({resource_type_camel: name})

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )

    if hub.tool.gcp.resource_prop_utils.properties_mismatch_resource_id(
        service_resource_type, resource_id, local_params
    ):
        result["comment"].append(
            hub.tool.gcp.comment_utils.properties_mismatch_resource_id_comment(
                gcp_service_resource_type, name
            )
        )

    if state_ctx.get("rerun_data"):
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            state_ctx,
            state_ctx.get("rerun_data"),
            service_resource_type,
        )

        if not handle_operation_ret["result"]:
            result["comment"] += handle_operation_ret["comment"]
            if handle_operation_ret.get("rerun_data"):
                result["rerun_data"] = handle_operation_ret["rerun_data"]
                if handle_operation_ret["rerun_data"].get("has_error", False):
                    result["result"] = False
            else:
                result["result"] = False

            return result

        resource_id = handle_operation_ret["resource_id"]

    if resource_id:
        old_get_ret = await hub_ref_exec.get(state_ctx, resource_id=resource_id)

        if not old_get_ret["result"] or (
            not old_get_ret["ret"]
            and (state_ctx.get("rerun_data") or get_resource_only_with_resource_id)
        ):
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        # long-running operation has succeeded - both update and create
        if state_ctx.get("rerun_data"):
            result["new_state"] = old_get_ret["ret"]
            result["old_state"] = state_ctx.get("rerun_data").get("old_state")

            # Check whether we need to run some code after a successful present call
            on_completion = state_ctx.get("rerun_data").get("on_completion")
            if on_completion:
                handler_method = hub.tool.gcp.hub_resolver.resolve(
                    on_completion["handler"]
                )
                ret = await handler_method(
                    state_ctx, result, on_completion.get("aux_ctx")
                )
                if ret:
                    result["comment"] += ret["comment"]

            if result["old_state"]:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.update_comment(
                        gcp_service_resource_type, name
                    )
                )
            else:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.create_comment(
                        gcp_service_resource_type, name
                    )
                )

            ctx_kwargs = {k: v for (k, v) in ctx.kwargs.items() if v is not None}
            diff = hub.tool.gcp.utils.compare_states(
                result["new_state"], ctx_kwargs, service_resource_type
            )
            values_mismatch: set = diff.get("relevant_changes")
            if values_mismatch:
                values_mismatch.discard("root['ctx']")
                if values_mismatch:
                    result["comment"].append(
                        f"Values mismatch between actual and desired states: {values_mismatch}"
                    )
            return result

        result["old_state"] = old_get_ret["ret"]
    elif not get_resource_only_with_resource_id:
        resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
            service_resource_type, local_params
        )

        if not resource_id:
            result["result"] = False
            result["comment"].append(
                f"Could not construct resource ID of {service_resource_type} from input arguments."
            )
            return result

        old_get_ret = await hub_ref_exec.get(state_ctx, resource_id=resource_id)

        if not old_get_ret["result"]:
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        if old_get_ret["ret"]:
            result["old_state"] = old_get_ret["ret"]

    state_ctx["wrapper_result"] = result
    return await ctx.func(*ctx.args, **{**ctx.kwargs, "resource_id": resource_id})


async def post_present(hub, ctx):
    r"""Evaluate the changes between old_state and new_state after present function call.

    This method compares both old_state and new_state and
    clears them from the fields that are expected to have changed,
    so they get ignored to avoid infinite reconciliation.

    Args:
        hub: The redistributed pop central hub. The root of the namespace that pop operates on.
        ctx: Invocation context for this command.

    Returns:
        The changes with removed exclude paths from them.

    """
    if ctx.ret and "old_state" in ctx.ret and "new_state" in ctx.ret:
        old_state = ctx.ret.get("old_state")
        new_state = ctx.ret.get("new_state")

        state_ctx = ctx.kwargs.get("ctx")
        resource_state_path = state_ctx.get("tag").split("_|")[0]
        service_resource_type = (
            hub.tool.gcp.resource_prop_utils.get_service_resource_type(
                resource_state_path
            )
        )

        changes = hub.tool.gcp.utils.compare_changes(
            old_state, new_state, service_resource_type
        )
        ctx.ret["changes"] = changes


async def call_absent(hub, ctx):
    r"""Wrapper for absent function.

    This method handles the parameters given in the ctx and deletes a resource only if
    the parameters are valid and the resource can be found using them.

    As with the call_present method, here in call_absent we also rely on having rerun_data in order to decide
    whether delete API call needs to be done or we just have to reconcile and wait until for the delete
    operation to complete.

    Steps:
        1. Construct resource_id if it is missing and get_resource_only_with_resource_id flag is NOT set.
            If resource_id is NOT provided and cannot be constructed, then it is directly assumed that the
            resource does not exist in the Cloud. We then return then result["resutl"]=True and result["comment"]
            that the resource is already absent.
        2. Using resource_id check for resource existence by getting the resource current state from the Cloud.
            If the resource is found then delete it or return already absent comment.

    Args:
        hub: The redistributed pop central hub. The root of the namespace that pop operates on.
        ctx: Invocation context for this command.


    Returns: The result of a resource deletion state.

    """
    state_ctx = ctx.kwargs.get("ctx")
    assert state_ctx, f"state context is missing: {state_ctx}"

    gcp_service_resource_type = state_ctx.get("tag").split("_|")[0]
    service_resource_type = gcp_service_resource_type.replace("gcp.", "")
    resource_type_camel = hub.tool.gcp.case.camel(
        gcp_service_resource_type.split(".")[-1]
    )

    if gcp_service_resource_type not in LIST_RESOURCES_WITH_ABSENT_WRAPPER:
        return await ctx.func(*ctx.args, **ctx.kwargs)

    name = ctx.kwargs.get("name", None)

    result = {
        "comment": [],
        "old_state": state_ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "result": True,
    }

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )

    resource_id = ctx.kwargs.get("resource_id")

    if not resource_id and not get_resource_only_with_resource_id:
        project = hub.tool.gcp.utils.get_project_from_account(
            state_ctx, ctx.kwargs.get("project")
        )
        zone = ctx.kwargs.get("zone")
        region = ctx.kwargs.get("region")
        resource_id = (state_ctx.get("old_state") or {}).get(
            "resource_id"
        ) or hub.tool.gcp.resource_prop_utils.construct_resource_id(
            service_resource_type,
            {
                **locals(),
                resource_type_camel: name,
            },
        )

    if not resource_id and not state_ctx.get("rerun_data"):
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(
                gcp_service_resource_type, name
            )
        )
        return result

    if not state_ctx.get("rerun_data"):
        resource_path = service_resource_type.split(".")
        hub_ref_exec = hub.exec.gcp
        for resource_path_segment in resource_path:
            hub_ref_exec = hub_ref_exec[resource_path_segment]

        get_ret = await hub_ref_exec.get(state_ctx, resource_id=resource_id)

        if not get_ret["result"]:
            result["result"] = False
            result["comment"] += get_ret["comment"]
            return result

        if not get_ret["ret"]:
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.already_absent_comment(
                    gcp_service_resource_type, name
                )
            )
            return result

        result["old_state"] = get_ret["ret"]
    else:
        result["old_state"] = state_ctx["rerun_data"]["old_state"]

    name = result["old_state"].get("name", name)

    if state_ctx.get("test"):
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(
                gcp_service_resource_type, name
            )
        )
        return result

    if not state_ctx.get("rerun_data"):
        hub_ref_exec_gcp_api = hub.exec.gcp_api.client
        for resource_path_segment in resource_path:
            hub_ref_exec_gcp_api = hub_ref_exec_gcp_api[resource_path_segment]

        # First iteration; invoke resource's delete()
        delete_ret = await hub_ref_exec_gcp_api.delete(
            state_ctx, resource_id=resource_id, request_id=ctx.kwargs.get("request_id")
        )

        if not delete_ret.get(
            "result"
        ) or not hub.tool.gcp.operation_utils.is_operation(delete_ret.get("ret")):
            result["result"] = False
            result["comment"].append(
                f"Unexpected return value from {service_resource_type}.delete - {delete_ret}"
            )
            return result

        result["result"] = True
        result["comment"] += delete_ret["comment"]
        operation_id = delete_ret["ret"].get("selfLink")
        result["rerun_data"] = {
            "operation_id": hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                operation_id,
                hub.tool.gcp.operation_utils.get_operation_type(operation_id),
            ),
            "old_state": result["old_state"],
        }
        return result
    else:
        # delete() has been called on some previous iteration
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, state_ctx.get("rerun_data"), service_resource_type
        )
        if not handle_operation_ret["result"]:
            result["comment"] += handle_operation_ret["comment"]
            if handle_operation_ret.get("rerun_data"):
                result["rerun_data"] = handle_operation_ret["rerun_data"]
                if handle_operation_ret["rerun_data"].get("has_error", False):
                    result["result"] = False
            else:
                result["result"] = False

            return result

        result["comment"].append(
            hub.tool.gcp.comment_utils.delete_comment(gcp_service_resource_type, name)
        )
    return result



# VMware Idem Plugin
# Copyright (c) 2020-2022 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any
from typing import Dict

from google.auth.credentials import Credentials


def post_gather(hub, ctx) -> Dict[str, Any]:
    """
    Validate the return from a "gather" function
    """
    for profile, profile_ctx in ctx.ret.items():
        error_msg = f"{ctx.ref} profile formatted incorrectly: {profile}"
        assert "credentials" in profile_ctx, error_msg
        assert isinstance(profile_ctx["credentials"], Credentials), error_msg

    return ctx.ret
